import time
import signal
import csv
import os
import re
import json
import shutil
import subprocess
import argparse
import sched
import logging
import multiprocessing
from random import random
from string import ascii_lowercase
from datetime import datetime, timedelta
from sys import getsizeof

from deps.binary import Monitor
from deps.file_regex_monitor import FileRegexMonitor
from deps import env_vars


COUNTER_INTERVAL = 0
N_ALERTS_JSON = 0


def generate_message_from_size(event_size, base_event=None, add_new_line=False):
    char_size = getsizeof(getsizeof('A') - getsizeof(''))
    message_size = event_size - getsizeof(base_event)
    event_to_use = base_event + 'A' * (message_size//char_size)
    event_to_use = event_to_use + '\n' if add_new_line else event_to_use
    return event_to_use


def write_events_header(events_file):
    header = ['timestamp', 'seconds', 'epi']
    header.append('num_alerts')

    write_csv_file(events_file, header)


def write_csv_file(filepath, data):
    with open(filepath, 'a', newline='') as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(data)


def get_alert_json_lines():
    lines_counter = 0
    with open(env_vars.MANAGER_ALERTS_JSON_FILEPATH) as fp:
        for lines_counter, line in enumerate(fp):
            pass
    return lines_counter


def write_csv_events_row(event_file, interval_csv, epi, alerts):
    current_date = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
    write_csv_file(event_file, [current_date, interval_csv, epi, alerts])


def events_monitoring(event_file, max_iteration, epi, interval, total_events):
    start_time = datetime.now().replace(microsecond=0)
    scheduler = sched.scheduler(time.time, time.sleep)
    iteration = 0
    generated_events = 0
    n_alerts_json = 0

    while iteration < max_iteration:
        interval_csv = interval * iteration
        new_interval = (start_time + timedelta(seconds=interval_csv)).timestamp()
        generated_events += epi
        real_epi = 0 if generated_events > total_events else epi

        lines_counter = get_alert_json_lines()
        new_alerts = lines_counter - n_alerts_json
        n_alerts_json = lines_counter

        scheduler.enterabs(new_interval, 0, write_csv_events_row, (event_file, interval_csv, real_epi, new_alerts))
        scheduler.run()
        iteration += 1


def syscheck_scan_callback(line):
    return 'File integrity monitoring scan ended.' in line


def syscheck_events_generator(n_files, file_update_size, interval_frequency, n_intervals,
                              directory, unexpected_failure):
    try:
        n_update_files_events = n_intervals if n_intervals <= 2 else (n_intervals - 1)

        for i in range(n_update_files_events):
            if unexpected_failure.value == 1:
                break
            for i in range(n_files):
                file_path = os.path.join(directory, f'file_{i}')
                with open(file_path, 'a') as f:
                    base_event = datetime.now().strftime("%b %d %H:%M:%S") + ' Local performance: ' + \
                                'Performance Testing - Syscheck:'
                    f.write(generate_message_from_size(file_update_size, base_event))
            regex_monitor = FileRegexMonitor(monitored_file=env_vars.MANAGER_LOGS_FILEPATH,
                                             callback=syscheck_scan_callback, timeout=max(interval_frequency*3, 120),
                                             only_new_events=True)

        if n_intervals > 2 and not unexpected_failure.is_set():
            shutil.rmtree(path=directory, ignore_errors=True)
            regex_monitor = FileRegexMonitor(monitored_file=env_vars.MANAGER_LOGS_FILEPATH,
                                             callback=syscheck_scan_callback,
                                             timeout=max(interval_frequency*3, 120), only_new_events=True)
    except Exception as e:
        unexpected_failure.value = 1


def logcollector_events_generator(logcollector_directory, file_events_per_second, event_size, logcollector_stress_time,
                                  unexpected_failure):
    try:
        for i in range(logcollector_stress_time):
            if unexpected_failure.value == 1:
                break
            for logcollector_file in os.listdir(logcollector_directory):
                for event in range(file_events_per_second):
                    with open(os.path.join(logcollector_directory, logcollector_file), 'a') as f:
                        base_event = datetime.now().strftime("%b %d %H:%M:%S") + ' Local performance: ' + \
                                    'Performance Testing - Logcollector:'
                        f.write(generate_message_from_size(event_size, base_event, True))
            time.sleep(1)
    except Exception as e:
        unexpected_failure.set()


def init_processes_monitoring(interval, statistics_path, data_unit, statistics_processes):
    wazuh_monitors = []
    for process in statistics_processes:
        for i, pid in enumerate(Monitor.get_process_pids(process)):
            p_name = process if i == 0 else f'{process}_child_{i}'
            monitor = Monitor(process_name=p_name, pid=pid, value_unit=data_unit, time_step=interval, version=None,
                              dst_dir=statistics_path)
            monitor.start()
            wazuh_monitors.append(monitor)
    return wazuh_monitors


def include_syscheck_configuration(configuration, directory, interval):
    new_configuration = configuration
    new_configuration += '<ossec_config>\n' + '<syscheck>\n' + \
                         '<disabled>no</disabled>\n' + \
                         f'<directories check_all="yes">{directory}</directories>\n' + \
                         '<alert_new_files>yes</alert_new_files>\n' + \
                         f'<scan_on_start>yes</scan_on_start>\n' + \
                         f'<frequency>{interval}</frequency>\n' + \
                         '</syscheck>\n' + '</ossec_config>\n'
    return new_configuration


def include_logcollector_configuration(configuration, directory, protocol):
    new_configuration = configuration
    new_configuration += '<ossec_config>\n' + '<localfile>\n' + \
                         '<log_format>syslog</log_format>\n' + \
                         f'<location>{directory}/*.log</location>\n' + \
                         '</localfile>' + \
                         '</ossec_config>\n'
    return new_configuration


def get_configuration():
    with open(MANAGER_CONFIGURATION_PATH) as f:
        return f.read()
