import os
import sys
import json
import signal
import shutil
import sys
import subprocess
import argparse
import multiprocessing
import threading
import time
import itertools
from datetime import datetime, timedelta

from deps.file_regex_monitor import FileRegexMonitor
from deps import env_vars
from deps import utils


def restore_environment():
    """Restore the original configuration and logs"""

    print("\n\nRestoring original configuration and logs...")

    TESTING_CASE_OVER = True

    # Restore original configuration
    shutil.copy(MANAGER_CONFIGURATION_FILEPATH_COPY, env_vars.MANAGER_CONFIGURATION_FILEPATH)

    # Restore original logs
    shutil.copy(MANAGER_LOGS_FILEPATH_COPY, env_vars.MANAGER_LOGS_FILEPATH)

    # Restore original alerts
    shutil.copy(MANAGER_ALERTS_JSON_COPY, env_vars.MANAGER_ALERTS_JSON_FILEPATH)
    shutil.copy(MANAGER_ALERTS_COPY, env_vars.MANAGER_ALERTS_FILEPATH)

    # Remove custom ruleset
    os.remove(CUSTOM_RULES_PATH)
    os.remove(CUSTOM_DECODERS_PATH)

    # Restart the manager
    time.sleep(5)


def SignalHandler_SIGINT(SignalNumber, Frame):
    """Handle the SIGINT signal"""
    global STRESS_PROCESSES
    global FOOTPRINT_MONITORS
    global ALERTS_MONITORING
    global ABORTED_TESTING
    global RESTARTING_MANAGER

    # Stop stress processes
    for process in STRESS_PROCESSES.values():
        process.terminate()

    # Stop alerts monitoring
    try:
        if ALERTS_MONITORING:
            if ALERTS_MONITORING.is_alive():
                print(ALERTS_MONITORING)
                ALERTS_MONITORING.terminate()
    except Exception as e:
        pass

    # Stop monitors
    for monitor in FOOTPRINT_MONITORS:
        monitor.shutdown()

    # Change loading animation message to aborted
    UNEXPECTED_FAILURE = 'ABORTED'
    ABORTED_TESTING = True

    if not RESTARTING_MANAGER:
        sys.exit(1)


def loading_animation(test_case):
    """Print a loading animation"""
    global TESTING_CASE_OVER
    for c in itertools.cycle(['|', '/', '-', '\\']):
        if TESTING_CASE_OVER:
            break
        print(f'\r Running {test_case}: ' + c, end='', flush=True)
        time.sleep(0.1)

    if UNEXPECTED_FAILURE:
        print(f'\r Running {test_case}: ' + f'Unexpected error\n', end='', flush=True)
    elif ABORTED_TESTING:
        print(f'\r Running {test_case}: ' + f'Aborted\n', end='', flush=True)
    else:
        print(f'\r Running {test_case}: ' + 'Done!\n', end='', flush=True)


def get_parameters():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-t', '--test-cases', metavar='<test-cases-file>', type=str,
                            help='Test case JSON file', required=True, default=None, dest='tests_cases_file')
    args = arg_parser.parse_args()
    test_cases = {}

    if os.path.exists(args.tests_cases_file):
        with open(args.tests_cases_file) as f:
            test_cases = json.load(f)

    return test_cases


# Register SIGINT handler
signal.signal(signal.SIGINT, SignalHandler_SIGINT)

MANAGER_LOGS_FILEPATH_COPY = f"{env_vars.MANAGER_LOGS_FILEPATH}.copy"
MANAGER_CONFIGURATION_FILEPATH_COPY = f"{env_vars.MANAGER_CONFIGURATION_FILEPATH}.copy"
MANAGER_ALERTS_JSON_COPY = f"{env_vars.MANAGER_ALERTS_JSON_FILEPATH}.copy"
MANAGER_ALERTS_COPY = f"{env_vars.MANAGER_ALERTS_FILEPATH}.copy"

ALERTS_CSV_FILENAME = 'alerts.csv'
LOGCOLLECTOR_TESTING_DIRECTORY = '/tmp/testing-logcollector'
SYSCHECK_TESTING_DIRECTORY = '/tmp/testing-syscheck'

STRESS_PROCESSES = {}
FOOTPRINT_MONITORS = []
ALERTS_MONITORING = None

TESTING_CASE_OVER = False
UNEXPECTED_FAILURE = False
ABORTED_TESTING = False
RESTARTING_MANAGER = False

# Define custom rules and decoders
CUSTOM_RULES = os.path.join('configuration', 'performance_rules.xml')
CUSTOM_DECODERS = os.path.join('configuration', 'performance_decoders.xml')
CUSTOM_RULES_PATH = os.path.join(env_vars.MANAGER_RULES_DIRECTORY, 'performance_rules.xml')
CUSTOM_DECODERS_PATH = os.path.join(env_vars.MANAGER_DECODERS_DIRECTORY, 'performance_decoders.xml')


def main():
    global TESTING_CASE_OVER
    global UNEXPECTED_FAILURE
    global STRESS_PROCESSES
    global FOOTPRINT_MONITORS
    global ALERTS_MONITORING
    global RESTARTING_MANAGER

    # Get parameters
    tests_cases = get_parameters()

    # Create results directory
    result_directory = datetime.now().strftime("%Y-%m-%d_%H%M%S")

    if os.path.exists(result_directory):
        input(f"Results directory {result_directory} already exists. Press enter to continue or CTRL+C to abort.")
        shutil.rmtree(path=result_directory, ignore_errors=True)

    os.mkdir(path=result_directory)

    # Backup original logs
    shutil.copy(env_vars.MANAGER_LOGS_FILEPATH, MANAGER_LOGS_FILEPATH_COPY)

    # Backup original logs
    shutil.copy(env_vars.MANAGER_LOGS_FILEPATH, MANAGER_LOGS_FILEPATH_COPY)

    shutil.copy(CUSTOM_RULES, CUSTOM_RULES_PATH)
    shutil.copy(CUSTOM_DECODERS, CUSTOM_DECODERS_PATH)

    # Stop the manager
    process = subprocess.Popen(env_vars.MANAGER_STOP_COMMAND.split(), stdout=subprocess.PIPE)

    # Backup alerts
    if os.path.exists(env_vars.MANAGER_ALERTS_JSON_FILEPATH):
        shutil.copy(env_vars.MANAGER_ALERTS_JSON_FILEPATH, MANAGER_ALERTS_JSON_COPY)
    if os.path.exists(env_vars.MANAGER_ALERTS_FILEPATH):
        shutil.copy(env_vars.MANAGER_ALERTS_FILEPATH, MANAGER_ALERTS_COPY)

    # Truncate logs
    with open(env_vars.MANAGER_LOGS_FILEPATH, 'w') as log_file:
        log_file.write('')

    # Truncate alerts
    with open(env_vars.MANAGER_ALERTS_JSON_FILEPATH, 'w') as alerts_json_file:
        alerts_json_file.write('')

    # Manager configuration backup
    shutil.copy(env_vars.MANAGER_CONFIGURATION_FILEPATH, MANAGER_CONFIGURATION_FILEPATH_COPY)

    try:
        print("----------------------------")
        for module_name, module_tests in tests_cases.items():
            print(f"\nRunning module: {module_name}")
            print("----------------------------")
            module_path = os.path.join(result_directory, module_name)
            os.mkdir(module_path)
            for case in module_tests:
                UNEXPECTED_FAILURE = False
                try:
                    loading_thread = threading.Thread(target=loading_animation, args=(case['name'],))
                    loading_thread.start()

                    # Create case evidences directory
                    case_evidences_path = os.path.join(module_path, case['name'])
                    shutil.rmtree(case_evidences_path, ignore_errors=True)
                    os.mkdir(case_evidences_path)

                    alerts_csv_file = os.path.join(case_evidences_path, ALERTS_CSV_FILENAME)

                    # Create testing directories
                    if not os.path.isdir(LOGCOLLECTOR_TESTING_DIRECTORY):
                        os.mkdir(LOGCOLLECTOR_TESTING_DIRECTORY)
                    if not os.path.isdir(SYSCHECK_TESTING_DIRECTORY):
                        os.mkdir(SYSCHECK_TESTING_DIRECTORY)

                    # Gather test case parameters
                    module = case.get('module', ['logcollector'])

                    event_size_logcollector = case.get('event_size_logcollector', 0)
                    logcollector_stress_time = case.get('logcollector_stress_time', 0)
                    logcollector_n_files = case.get('logcollector_n_files', 0)
                    logcollector_feps = case.get('logcollector_feps', 0)

                    file_update_size = case.get('file_update_size', 0)
                    syscheck_n_files = case.get('syscheck_n_files', 0)
                    interval_frequency = case.get('interval', 0)
                    n_intervals = case.get('n_intervals', 0)

                    # Create files
                    if 'logcollector' in module:
                        for i in range(logcollector_n_files):
                            with open(os.path.join(LOGCOLLECTOR_TESTING_DIRECTORY, f"file{i}.log"), 'w') as f:
                                f.write('\n')

                    # Configure the manager
                    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                           os.path.join('configuration',
                                                        'plain_configuration.conf'))) as configuration_file:
                        new_configuration = configuration_file.read()

                    if 'syscheck' in module:
                        new_configuration = utils.include_syscheck_configuration(new_configuration,
                                                                                 SYSCHECK_TESTING_DIRECTORY,
                                                                                 str(interval_frequency))
                    if 'logcollector' in module:
                        new_configuration = utils.include_logcollector_configuration(new_configuration,
                                                                                     LOGCOLLECTOR_TESTING_DIRECTORY,
                                                                                     'syslog')

                    with open(env_vars.MANAGER_CONFIGURATION_FILEPATH, 'w') as f:
                        f.write(new_configuration)

                    # Restart the manager
                    RESTARTING_MANAGER = True
                    process = subprocess.check_call(env_vars.MANAGER_RESTART_COMMAND.split(), stdout=subprocess.DEVNULL,
                                                    stderr=subprocess.DEVNULL)
                    RESTARTING_MANAGER = False

                    if ABORTED_TESTING:
                        raise ValueError("Aborted testing process")

                    process_to_monitor = [env_vars.ANALYSISD_DAEMON]

                    # Init footprint
                    if 'logcollector' in module:
                        process_to_monitor.append(env_vars.LOGCOLLECTORD_DAEMON)

                    if 'syscheck' in module:
                        process_to_monitor.append(env_vars.SYSCHECKD_DAEMON)

                    FOOTPRINT_MONITORS = utils.init_processes_monitoring(1, case_evidences_path, 'KB',
                                                                         process_to_monitor)

                    # Init modules stress
                    unexpected_error = multiprocessing.Value('i', 0)

                    if 'syscheck' in module:
                        regex_monitor = FileRegexMonitor(monitored_file=env_vars.MANAGER_LOGS_FILEPATH,
                                                         callback=utils.syscheck_scan_callback,
                                                         timeout=interval_frequency * 2)

                        process = multiprocessing.Process(target=utils.syscheck_events_generator,
                                                          args=(syscheck_n_files,
                                                                file_update_size, interval_frequency, n_intervals,
                                                                SYSCHECK_TESTING_DIRECTORY, unexpected_error,))

                        process.start()
                        STRESS_PROCESSES['syscheck_stress'] = process

                    if 'logcollector' in module:
                        process = multiprocessing.Process(target=utils.logcollector_events_generator,
                                                          args=(LOGCOLLECTOR_TESTING_DIRECTORY, logcollector_feps,
                                                                event_size_logcollector, logcollector_stress_time,
                                                                unexpected_error,))
                        process.start()
                        STRESS_PROCESSES['logcollector_stress'] = process

                    utils.write_events_header(alerts_csv_file)

                    max_time = max(logcollector_stress_time, (n_intervals)*interval_frequency)

                    total_events = (logcollector_n_files * logcollector_feps) * \
                        logcollector_stress_time + syscheck_n_files * (n_intervals)

                    total_epi = int(total_events / max_time)

                    ALERTS_MONITORING = multiprocessing.Process(target=utils.events_monitoring,
                                                                args=(alerts_csv_file, max_time*10, total_epi,
                                                                      1, total_events,))
                    ALERTS_MONITORING.start()

                    # Wait modules stress
                    for process in STRESS_PROCESSES.values():
                        process.join()

                    if unexpected_error.value == 0:
                        extra_interval = 10
                        time.sleep(extra_interval)
                    else:
                        UNEXPECTED_FAILURE = True

                    ALERTS_MONITORING.terminate()

                    # Stop monitors
                    for monitor in FOOTPRINT_MONITORS:
                        monitor.shutdown()

                    # Gather evidences
                    shutil.copy(env_vars.MANAGER_LOGS_FILEPATH, os.path.join(case_evidences_path, 'ossec.log'))

                except (TimeoutError, OSError) as e:
                    if not ABORTED_TESTING:
                        print(f"Unexpected error: {e}")
                    else:
                        print("Aborted testing process")

                finally:
                    TESTING_CASE_OVER = True
                    time.sleep(1)
                    TESTING_CASE_OVER = False

                    for process in STRESS_PROCESSES.values():
                        if process.is_alive():
                            process.terminate()

                    if ALERTS_MONITORING and ALERTS_MONITORING.is_alive():
                        ALERTS_MONITORING.terminate()

                    for monitor in FOOTPRINT_MONITORS:
                        monitor.shutdown()

                    with open(env_vars.MANAGER_LOGS_FILEPATH, 'w') as log_file:
                        log_file.write('')

                    with open(env_vars.MANAGER_ALERTS_JSON_FILEPATH, 'w') as log_file:
                        log_file.write('')

                    shutil.rmtree(path=LOGCOLLECTOR_TESTING_DIRECTORY, ignore_errors=True)
                    shutil.rmtree(path=SYSCHECK_TESTING_DIRECTORY, ignore_errors=True)

                    if ABORTED_TESTING:
                        print("Aborting testing process")
                        sys.exit()

    finally:
        print("Restoring environment")
        restore_environment()
        print("Restarting the manager")
        process = subprocess.Popen(env_vars.MANAGER_RESTART_COMMAND.split(), stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL)
        process.communicate()
        print("Manager restarted")


if __name__ == '__main__':
    main()
